/******************************************************************************
 *
 * The main function is divided into three pieces. The first piece defines all
 * of the variables that are needed throughout the entire program. The second
 * piece is the water flow portion. This solves the system for the water depth
 * through and iterative process that stops when the water depth has come to a
 * sufficient equilibrium. The third section of this code solves for the height
 * of the water surface by breaking the system into a series of tridiagonal 
 * systems and solving each one independently. The code repeats the water flow
 * and sediment flow sections of the code until a sufficient amount of material
 * has been eroded.
 *
 *****************************************************************************/

/* Needed System Files */
#include <mpi.h>
#include <sstream>

/* User Defined Files */
#include "getXY.h"
#include "constants.h"
#include "read_data.h"
#include "write_data.h"
#include "triDiagonalSolver.h"

using namespace std;

int main ( int argc, char **argv )
{
  // Define the variables that hold the important information about the system.
  int index;			
  double target, volume;
  double H[LENGTH][WIDTH] = {};
  double h[LENGTH][WIDTH] = {};

  // Get the name of the file with all the data
  string file_name = argv[1];
  // Get the path of the directory to save data to
  string path_name = argv[2];

  // For path naming
  std::ostringstream oss;

  // Define a variable to dictate when the process is finished
  bool done = false;

  // Define a variable to store the current percent eroded 
  double cur_percent = 0;

  // Define general loop variables
  int i, j;

  /*****************************************************************************
   *
   * START: Water Flow Variables
   *   The following variables are all used by the water flow section
   * of the proceeding code:
   *
   ****************************************************************************/

  // Variables for velocities at half grid points:
  double U[LENGTH][WIDTH-1] = {};	// U[i][j] ~ u(i + 0.5, j)
  double V[LENGTH][WIDTH-1] = {};	// V[i][j] ~ v(i, j + 0.5)
  
  // A second array to store (intermediate) water depth values
  double hn[LENGTH][WIDTH] = {};

  // Variables to make the calculations of U and V simplier
  int subI;
  double Hx, Hy, denominator;

  // Variables to aid in the actual calculation of the new water depth
  bool equilibrium;
  double fluxRight, fluxLeft, fluxUp, fluxDown;
  
  /*****************************************************************************
   * END: Water Flow Variables
   ****************************************************************************/
  
  /*****************************************************************************
   *
   * START: Sediment Flow Variables
   *   The following variables are all used by the sediment flow section
   * of the proceeding code:
   *
   ****************************************************************************/

  // Variables to ensure stability when eroded
  double lambda;
  double dt, m = 0;
  double XY[2][21];
  
  // Variables used for solving the various tri-diagonal systems
  // Variables to solve in the y, y=x, and y=-x directions
  double Sol[WIDTH+1];
  
  double Ay [WIDTH+1];
  double By [WIDTH+1];
  double Cy [WIDTH+1];
  double by [WIDTH+1];

  // Majority of the boundary conditions never change
  By[0] = 0;
  Ay[0] = 1;
  Cy[0] = 0;

  By[WIDTH] = -1;
  Ay[WIDTH] = 1;
  Cy[WIDTH] = 0;
  by[WIDTH] = 0;

  // Variables to solve in the x direction
  double x [LENGTH-1];
  double y [LENGTH-1];

  double Ax [LENGTH];
  double Bx [LENGTH];
  double Cx [LENGTH];
  double bx1 [LENGTH];
  double bx2 [LENGTH-1] = {};

  // Arrays used for solving/storing the tridiagonal systems
  double phi1[LENGTH][WIDTH];
  double phi2[LENGTH][WIDTH];

  // Used for the diagonal systems to keep track of which points are 
  // being solved for
  int I[WIDTH];
  
  /*****************************************************************************
   * END: Sediment Flow Variables
   ****************************************************************************/

  /*****************************************************************************
   *
   * Start the MPI implementaion.
   *
   ****************************************************************************/

  // Define variables specifically needed for the MPI implementation
  bool equal;
  bool master;
  int rank, size;
  int row, column;
  int row_chunks, col_chunks;
  int sed_row_lower, sed_row_upper;
  double col_data_send[LENGTH/2 + 1] = {};	// Maximum Value Needed
  double col_data_receive[LENGTH/2 + 1] = {};	// Maximum Value Needed
  int process_left, process_right, process_up, process_down; 
  int water_row_lower, water_row_upper, water_col_lower, water_col_upper;
  // Start the Parallelism
  MPI_Init( &argc, &argv );
  MPI_Comm_size( MPI_COMM_WORLD, &size );
  MPI_Comm_rank( MPI_COMM_WORLD, &rank );

  // Determine whether or not a thread is The Master
  // This should make the implementation a little easier
  master = ( rank == 0 );

  // Have The Master determine how to divide the domain for the
  // water flow portion
  if ( master )
  {
    // Divide the domain as evenly as possible
    i = (int) sqrt((double) size);
    for ( j=1; j < (i+1); j++ )
    {
      if ( size % j == 0 )
      {
        row_chunks = size/j;
        col_chunks = j;
      }
    }

    // Get the initial data
    read_data( file_name, &index, &target, &volume, H, h );
  } // End The Master's work

  // Have The Master broadcast the break up of the domain. The work
  // done by each process is determined by its rank
  MPI_Bcast( &row_chunks, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &col_chunks, 1, MPI_INT, 0, MPI_COMM_WORLD );

  // Now each process has the row_chunks and col_chunks, and thus
  // can figure out its section of work.
  row = rank % row_chunks;
  column = rank / row_chunks;
  // Work for Water Flow Portion (Grid Tiling)
  // Rows:
  water_row_lower = row*(LENGTH/row_chunks);
  water_row_upper = (row+1)*(LENGTH/row_chunks);
  if ( (row+1) == row_chunks ) { water_row_upper = LENGTH; }
  // Columns:
  water_col_lower = column*(WIDTH/col_chunks);
  water_col_upper = (column+1)*(WIDTH/col_chunks);
  if ( (column+1) == col_chunks ) { water_col_upper = WIDTH-1; }

  // Get the amount of work responsible for the sediment flow
  // section. This division can only be done amongst the rows.
  sed_row_lower = rank*(LENGTH/size);
  sed_row_upper = (rank+1)*(LENGTH/size);
  if ( (rank+1) == size ) { sed_row_upper = LENGTH; }

  // To make the sending of data between processes easier, each process
  // will now figure out its neighboring processes.
  // ===========================================================================
  // NOTE: These neighbors are only for the division of the data for the
  // grid tiling for the water flow section. The neighbors for the sediment
  // flow section can be found very easily using modulor arithmetic on the
  // rank of the process.
  // ===========================================================================
  process_up = (rank + row_chunks) % size;
  process_down = ( rank + (col_chunks-1)*row_chunks ) % size;
  process_right = column*row_chunks + ( (rank+1) % row_chunks );
  process_left = column*row_chunks + ( (rank+row_chunks-1) % row_chunks ); 

  // ===========================================================================
  // START: Erosion
  //   Now we start the erosion process with the equilibration of the water. We
  // repeat these steps until we are done.
  // ===========================================================================
  while ( !done )
  {
    /***************************************************************************
     *
     * START: Water Flow Process
     *   This section uses an upwind scheme to calculate the depth
     *  of water over a period of time, given by TIME_STEP. The water
     *  depth is recalculated until it reaches an equilibrium.
     *
     **************************************************************************/
     
    // The values in H will not change in this section, and every process will
    // need all the values of H for the sediment flow section. To simply things,
    // have The Master broadcast the value of H
    MPI_Bcast( H, LENGTH*WIDTH, MPI_DOUBLE, 0, MPI_COMM_WORLD );

    // To minimize the number of messages sent, have The Master broadcast
    // all of the h matrix. Each process will only use what they need
    MPI_Bcast( h, LENGTH*WIDTH, MPI_DOUBLE, 0, MPI_COMM_WORLD );
    
    // The first step is to calculate the velocities at the half grid points
    // Have each process calculate only the values they will need. 
    for ( i=water_row_lower-1; i < water_row_upper; i++ )
    {
      // This subI variable is for when water_row_lower = 0 and we need data
      // from the "left" of this point. 
      // This gives us the value of i ( and thus the value of U[i][j] ) that
      // will be needed later. i.e. The values to the left of the 
      // water_row_lower row. 
      subI = LEFT( i + 1 );
      for ( j=max(0, water_col_lower-1); j < water_col_upper; j++ )
      {
        // Calculate the Velocity at (i, j+0.5)
        Hx = ( H[RIGHT(subI)][j] + H[RIGHT(subI)][UP(j)] - H[LEFT(subI)][j] - H[LEFT(subI)][UP(j)] )/(4*DX);
        Hy = ( H[subI][UP(j)] - H[subI][j] )/DY;

        // Calculate the denominator
        denominator = pow( Hx*Hx + Hy*Hy, 0.25 );
        // The final result
        V[subI][j] = -1.0*MANNING*(Hy/denominator);

        // Calculate the Velocity at (i+0.5, j)
        Hx = ( H[RIGHT(subI)][j] - H[subI][j] )/DX;
        Hy = ( H[subI][UP(j)] + H[RIGHT(subI)][UP(j)] - H[subI][DOWN(j)] - H[RIGHT(subI)][DOWN(j)] )/(4*DY);

        // Calculate the denominator
        denominator = pow( Hx*Hx + Hy*Hy, 0.25 );
        // The final result
        U[subI][j] = -1.0*MANNING*(Hx/denominator);
      }
    }

    // Now we let the water move until it reaches an equilibrium
    equilibrium = false; // Make sure it runs at least once

    while ( !equilibrium )
    {
      // Assume that equilibrium will be reached this run
      equal = true;

      // Copy the needed values into the dummy hn array
      for ( i=water_row_lower; i < water_row_upper; i++ )
      {
        for ( j=water_col_lower; j < water_col_upper+1; j++ )
        {
          hn[i][j] = h[i][j];
        }
      }

      // Data at the boundaries of each process needs to be passed around.
      // i.e. Processes need the most recent data at their ghost points.
      // Pass data to the process on the right
      MPI_Request req1, req2;
      // Start a non blocking receive to collect data
      MPI_Irecv( &(hn[LEFT(water_row_lower)][water_col_lower]), (water_col_upper - water_col_lower), MPI_DOUBLE, process_left, 0, MPI_COMM_WORLD, &req1 );
      // Send the data that needs to be send to the right
      MPI_Send( &(h[water_row_upper-1][water_col_lower]), (water_col_upper - water_col_lower), MPI_DOUBLE, process_right, 0, MPI_COMM_WORLD );
      // Wait to receive your data before continuing
      MPI_Wait( &req1, MPI_STATUS_IGNORE );  

      // Send data to the process on the left
      MPI_Irecv( &(hn[RIGHT(water_row_upper-1)][water_col_lower]), (water_col_upper - water_col_lower), MPI_DOUBLE, process_right, 1, MPI_COMM_WORLD, &req2 );
      MPI_Send( &(h[water_row_lower][water_col_lower]), (water_col_upper - water_col_lower), MPI_DOUBLE, process_left, 1, MPI_COMM_WORLD );
      MPI_Wait( &req2, MPI_STATUS_IGNORE );

      // Send data above and below, if there are processes there
      if ( col_chunks > 1 )
      { 
        MPI_Request req3, req4;
        // Now send data to the process above.
        // If no such process exists, the receiving process 
        // just does not use the given data
        
        // Put the data that needs to be sent in an array for easy sending
        for ( int p=0; p < (water_row_upper - water_row_lower); p++ )
          col_data_send[p] = h[water_row_lower+p][water_col_upper-1];

        // Start a non blocking receive of data
        MPI_Irecv( col_data_receive, (water_row_upper - water_row_lower), MPI_DOUBLE, process_down, 2, MPI_COMM_WORLD, &req3 );
        // Send the data
        MPI_Send( col_data_send, (water_row_upper - water_row_lower), MPI_DOUBLE, process_up, 2, MPI_COMM_WORLD );
        MPI_Wait( &req3, MPI_STATUS_IGNORE );
        // Only store the data if this process was suppose to get data
        if ( process_down < rank )
        {
          for ( int p=0; p < (water_row_upper - water_row_lower); p++ )
            hn[water_row_lower+p][DOWN(water_col_lower)] = col_data_receive[p];
        }

        // Make sure all processes are on the same page
        MPI_Barrier( MPI_COMM_WORLD );

        // Now send data to the process below.
        // If no such process exists, the receiving process
        // just does not use the given data
        
        // Put the data that needs to be send in an array for easy sending
        for ( int p=0; p < (water_row_upper - water_row_lower); p++ )
          col_data_send[p] = h[water_row_lower+p][water_col_lower];

        // Receive Data 
        MPI_Irecv( col_data_receive, (water_row_upper - water_row_lower), MPI_DOUBLE, process_up, 3, MPI_COMM_WORLD, &req4 );
        // Send the data
        MPI_Send( col_data_send, (water_row_upper - water_row_lower), MPI_DOUBLE, process_down, 3, MPI_COMM_WORLD );
        MPI_Wait( &req4, MPI_STATUS_IGNORE );
        if ( process_up > rank )
        { 
          for ( int p=0; p < (water_row_upper - water_row_lower); p++ )
            hn[water_row_lower+p][water_col_upper] = col_data_receive[p];
        }
      }
      // Now every process should have the current data 
      // it needs to continue

      // Make sure all processes are on the same page
      MPI_Barrier( MPI_COMM_WORLD );

      // Loop through (almost) every point in the array and calculate the 
      // new water depth value.
      for ( i=water_row_lower; i < water_row_upper; i++ )
      {
        for ( j=water_col_lower; j < water_col_upper; j++ )
        {
          // Calculate the flux above this point
          fluxUp = ( hn[i][UP(j)] + hn[i][j] )/2;
          fluxUp = V[i][j]*pow( fluxUp, 2./3 );

          // Determine the direction the water is moving
          // and adjust accordingly
          if ( fluxUp > 0 )	fluxUp *= hn[i][j];
          else			fluxUp *= hn[i][UP(j)];

          // Calculate the flux below this point
          fluxDown = ( hn[i][j] + hn[i][DOWN(j)] )/2;
          fluxDown = V[i][DOWN(j)]*pow( fluxDown, 2./3 );

          // Determine the direction the water is moving
          // and adjust accordingly
          if ( fluxDown > 0 )	fluxDown *= hn[i][DOWN(j)];
          else			fluxDown *= hn[i][j];

          // Calculate the flux to the right of this point
          fluxRight = ( hn[RIGHT(i)][j] + hn[i][j] )/2;
          fluxRight = U[i][j]*pow( fluxRight, 2./3 );

          // Determine the direction the water is moving
          // and adjust accordingly
          if ( fluxRight > 0 )	fluxRight *= hn[i][j];
          else			fluxRight *= hn[RIGHT(i)][j];

          // Calculate the flux to the left of this point
          fluxLeft = ( hn[LEFT(i)][j] + hn[i][j] )/2;
          fluxLeft = U[LEFT(i)][j]*pow( fluxLeft, 2./3 );

          // Determine the direction the water is moving
          // and adjust accordingly
          if ( fluxLeft > 0 )	fluxLeft *= hn[LEFT(i)][j];
          else			fluxLeft *= hn[i][j];

          // Calculate the new water depth
          h[i][j] += ((TIME_STEP)/(SCALING*SCALING))*(RAINFALL - ((fluxRight - fluxLeft)/DX) - ((fluxUp - fluxDown)/DY) );

          // Check if the difference is small enough for
          // the desired convergence
          equal = equal && (fabs( h[i][j]-hn[i][j] ) < TOLERANCE); 
        }
      }

      // Now apply the boundary conditions

      // The top has a constant water depth of zero
      if ( water_col_upper == WIDTH-1 )
      {
        for ( i=water_row_lower; i < water_row_upper; i++ ) 
          { h[i][water_col_upper] = 0; }
      }

      // Determine whether or not we need to continue
      // Have The Master collect all the values of equilibrium 
      // then broadcast the result
      MPI_Allreduce( &equal, &equilibrium, 1, MPI::BOOL, MPI_LAND, MPI_COMM_WORLD );
    } // End Equilibrium While    

    // At this point, the water depth has reached an equilibrium.
    // Have The Master collect the water depth data and store it 
    // correctly in the original variable h. To do this, each process
    // will zero the hn array, put its data into that array, then use
    // the Reduce function to get the combined result
    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );
    // Move the data
    for ( i=water_row_lower; i < water_row_upper; i++ )
    {
      for ( j=water_col_lower; j < water_col_upper; j++ )
        hn[i][j] = h[i][j];
    }

    // Reduce the solution
    MPI_Allreduce( hn, h, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    /***************************************************************************
     * END: Water Flow Process
     **************************************************************************/

    /***************************************************************************
     *
     * START: Sediment Flow Process
     *   This section uses a Crank-Nicholson scheme to erode a 
     *  small amount of sediment from the surface. This is done by dividing
     *  work into a series of tridiagonal systems to solve. First the solution
     *  matrix/vector must be found, then it can be iteratively solved.
     *
     **************************************************************************/
    
    // To ensure stability, no more than 1 mm of sediment can be removed
    // from any one gridpoint during this process. To ensure this, we need to
    // do some calculations to find the time it will take to erode no more 
    // than a millimeter of sediment.
    m = 0;

    // =========================================================================
    // At every gridpoint (except where y=0) we need to calculate
    for ( i=water_row_lower; i < water_row_upper; i++ )
    {
      for ( j=water_col_lower+1; j < water_col_upper+1; j++ )
      {
        getXY( i, j, XY, H, h );
        // Now we have the necessary values, calculate the max value
        // Use Hx and Hy as temperary storage
        Hx = XY[0][13]*XY[0][5] - XY[0][14]*XY[0][6];
        Hy = XY[1][13]*XY[1][5] - XY[1][14]*XY[1][6];
        m = max(m, fabs( Hx + Hy ) );
      }
    }

    // Now we need to reduce the value of m, calculate the time step,
    // and broadcast the time step in the form of the variable lambda. 
    // To store the incoming values of m, use a spare double not need in 
    // this section: Hx.
    Hx = 0;
    MPI_Allreduce( &m, &Hx, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );
    // Calculate the time step
    dt = (0.001*DX)/(Hx*E);	
    lambda = (E*dt)/(2*DX);
   
    // Now that we have lambda, we can calculate the solution vector in
    // the resulting linear system. Clear needed arrays first.
    memset( phi1, 0, LENGTH*WIDTH*sizeof(double) );
    memset( phi2, 0, LENGTH*WIDTH*sizeof(double) );

    // We are going to have to perform these calculations multiple times to 
    // get the solution vector.
    //
    // To reduce the number of data being passed between processes, 
    // strategically solve at locations. 
    //
    // Solve for the first two pieces
    for ( i = sed_row_lower; i < sed_row_upper; i++ )
    {
      for ( j = 0; j < WIDTH; j ++ )
      {
        // Find were to solve
        if ( j == 0 ) { I[0] = i; }
        else { I[j] = LEFT( I[j-1] ); }

        // Get needed data
        getXY( I[j], j, XY, H, h );

        // Solve for the needed data
        // This is the system in the x-direction
        phi1[I[j]][j] = 0.5*( XY[1][17] - XY[1][20])*XY[0][0];
        phi1[I[j]][j] += (XY[0][13] + XY[0][16])*XY[0][5];
        phi1[I[j]][j] += (XY[0][14] + XY[0][19])*XY[0][6];
        phi1[I[j]][j] = H[I[j]][j] + lambda*phi1[I[j]][j];
      }
      // There is one extra piece of data that will be needed
      if ( i == sed_row_lower )
      {
        j = WIDTH-1;
        subI = LEFT( I[j] );
        getXY( subI, j, XY, H, h );
        
        phi1[subI][j] = 0.5*( XY[1][17] - XY[1][20])*XY[0][0];
        phi1[subI][j] += (XY[0][13] + XY[0][16])*XY[0][5];
        phi1[subI][j] += (XY[0][14] + XY[0][19])*XY[0][6];
        phi1[subI][j] = H[subI][j] + lambda*phi1[subI][j];
      } 
    } 

    // Each process solved for locations needed to solve the next direction
    for ( i = sed_row_lower; i < sed_row_upper; i ++ )
    {
      for ( j = 0; j < WIDTH; j ++ )
      {
        // Find were to solve
        if ( j == 0 ) { I[0] = i; }
        else { I[j] = LEFT( I[j-1] ); }
        
	// Get needed data
        getXY( I[j], j, XY, H, h );

        // Solve for the need data
        // This is the system in the y=-x direction
        phi2[I[j]][j] = ( XY[0][17] + XY[1][20] )*phi1[RIGHT(I[j])][DOWN(j)];
        phi2[I[j]][j] += ( XY[0][20] + XY[1][17] )*phi1[LEFT(I[j])][UP(j)];
        phi2[I[j]][j] = phi1[I[j]][j] - ( 0.25*lambda*phi2[I[j]][j] )/(DX); 
      }
    }
    
    // Now we need to merge data and distribute to finish solving
    // Combine the data
    MPI_Allreduce( phi2, phi1, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    // Clear phi2
    memset( phi2, 0, LENGTH*WIDTH*sizeof(double) );
   
    // Now solve in the last two directions, again strategically.
    for ( i = sed_row_lower; i < sed_row_upper; i ++ )
    {
      for ( j = 0; j < WIDTH; j++ )
      {
        getXY( i, j, XY, H, h );
        // Now solve for the next part of the solution vector
        // This is the y=x direction
        phi2[i][j] = ( XY[0][17] + XY[1][17] )*phi1[RIGHT(i)][UP(j)];
        phi2[i][j] += ( XY[0][20] + XY[1][20] )*phi1[LEFT(i)][DOWN(j)];
        phi2[i][j] = phi1[i][j] + ( 0.25*lambda*phi2[i][j] )/(DX); 
      }
    }

    // Clear phi1
    memset( phi1, 0, LENGTH*WIDTH*sizeof(double) );
  
    // Perfom one last calculation  
    for ( i = sed_row_lower; i < sed_row_upper; i++ )
    {
      for ( j = 0; j < WIDTH; j++ )
      {
        getXY( i, j, XY, H, h );
        // Now solve for the last part of the solution vector
        // This system is in the y direction
        // Since this is the last calculation, we also add the 'constant' term
        
        // Constant Term:
	phi1[i][j] = XY[0][13]*XY[0][5] + XY[1][13]*XY[1][5];
	phi1[i][j] -= XY[0][14]*XY[0][6] + XY[1][14]*XY[1][6];

        // The Equation:
        phi1[i][j] += 0.5*(XY[0][17] - XY[0][20])*( (phi2[i][UP(j)] - phi2[i][j-1])/(2*DY) );
        phi1[i][j] += ( XY[1][13] + XY[1][16] )*( (phi2[i][UP(j)] - phi2[i][j])/DY );
        phi1[i][j] += ( XY[1][14] + XY[1][19] )*( (phi2[i][j] - phi2[i][j-1])/DY );
        
        // The final solution:
        phi1[i][j] = phi2[i][j] + lambda*phi1[i][j];
      }
    }

    // Now each process has the data it needs to solve for the new water level.
    // Go right into solving
    
    // Each process has a subset of data it is responsible for solving for
    // the results are going to be stored in the hn array, then reduced back
    // to the original H variable.
    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );

    // =========================================================================
    // The y Direction

    for ( i=sed_row_lower; i < sed_row_upper; i++ )
    {
      // Set up the 1D tri-diagonal system's boundary conditions
      by[0] = phi1[i][0]; 
      
      // Set up the rest of the tri-diagonal system
      for ( j=1; j < WIDTH; j ++ )
      {
        getXY( i, j, XY, H, h );
        // Now we can calculate the needed values:
        By[j] = (-0.5)*( XY[0][17] - XY[0][20] )/(2*DY);
      
        Cy[j] = (-1.0)*By[j]; // Cy uses the same calculation
        
        By[j] -= ( XY[1][14] + XY[1][19] )/DY;
        By[j] *= lambda;
        
        Cy[j] += ( XY[1][13] + XY[1][16] )/DY;
        Cy[j] *= lambda;

        Ay[j] = ( XY[1][14] + XY[1][19] - XY[1][13] - XY[1][16] )/DY;
        Ay[j] = 1 + lambda*Ay[j];
       
        by[j] = phi1[i][j];
      }

      // Solve the linear system
      triDiagonalSolver( Sol, By, Ay, Cy, by, WIDTH+1 );

      // Put the solution in the correct location in the hn array
      for ( j=0; j < WIDTH; j++ ) { hn[i][j] = Sol[j]; }
    }

    // Now collect the results, apply the boundary condition,
    // broadcast the new values of H, and repeat for the next direction
    MPI_Allreduce( hn, phi2, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );

    // =========================================================================
    // The y = x Direction

    for ( i=sed_row_lower; i < sed_row_upper; i++ )
    {
      by[0] = phi2[i][0];
 
      // Because this direction is diagonal, we need a way to track which
      // row the data is in.
      I[0] = i;

      for ( j=1; j < WIDTH; j++ )
      {
        I[j] = RIGHT(I[j-1]);
        getXY( I[j], j, XY, H, h );
        // Calculate the needed values
        By[j] = (0.25)*lambda*(XY[0][20] + XY[1][20])/DX;
        Cy[j] = (0.25)*lambda*(XY[0][17] + XY[1][17])/DX;
        Ay[j] = 1;

        by[j] = phi2[I[j]][j];
      }

      // Solve the linear system
      triDiagonalSolver( Sol, By, Ay, Cy, by, WIDTH+1 );

      // Put the solution in the correct location in the hn array
      for ( j=0; j < WIDTH; j++ ) { hn[(I[j])][j] = Sol[j]; }
    }

    // Now collect the results, apply the boundary condition,
    // broadcast the new values of f, and repeat for the next direction
    MPI_Allreduce( hn, phi2, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );

    // =========================================================================
    // The y = -x Direction

    for ( i=sed_row_lower; i < sed_row_upper; i++ )
    {
      by[0] = phi2[i][0]; 

      // Again, since this is a diagonal direction, this is needed to
      // keep track of the row position
      I[0] = i;

      for ( j=1; j < WIDTH; j++ )
      {
        I[j] = LEFT(I[j-1]);
        
        getXY( I[j], j, XY, H, h );
        // Calculate the needed values
        By[j] = (-0.25)*lambda*( XY[0][17] + XY[1][20] )/DX;
        Cy[j] = (-0.25)*lambda*( XY[0][20] + XY[1][17] )/DX;
        Ay[j] = 1;

        by[j] = phi2[I[j]][j];
      }
      
      // Solve the linear system
      triDiagonalSolver( Sol, By, Ay, Cy, by, WIDTH+1 );

      // Put the solution in the correct location in the hn array
      for ( j=0; j < WIDTH; j++ ) { hn[(I[j])][j] = Sol[j]; }
    }

    // Now collect the results, apply the boundary condition,
    // broadcast the new values of phi2, and repeat for the next direction
    MPI_Allreduce( hn, phi2, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );

    //==========================================================================
    // The x Direction
    
    // This direction is a little different as we are moving across the surface
    // and not up
    for ( j = rank+1; j < WIDTH; j += size )
    {
      // Set up the 1D tri-diagonal system's boundary conditions
      // The boundary conditions for the tri-diagonal systems are
      // different due to the periodic conditions of the system.
      // Find all needed values as normal, without regard for the boundary
      for ( i=0; i < LENGTH; i ++ )
      {
        getXY( i, j, XY, H, h );
        // Now we can calculate the needed values:
        Bx[i] = (-0.5)*( XY[1][17] - XY[1][20] )/(2*DX);

        Cx[i] = (-1.0)*Bx[i]; // Cx uses the same calculation

        Bx[i] -= (XY[0][14] + XY[0][19])/DX;
        Bx[i] *= lambda;

        Cx[i] += (XY[0][13] + XY[0][16])/DX;
        Cx[i] *= lambda;       

        Ax[i] = ( XY[0][14] + XY[0][19] - XY[0][13] - XY[0][16] )/DX;
        Ax[i] = 1 + lambda*Ax[i];

        bx1[i] = phi2[i][j];
      }
      // Create the other solution vector
      bx2[0] = (-1.0)*Bx[0];
      bx2[LENGTH-2] = (-1.0)*Cx[LENGTH-2];

      // Since the boundary conditions are period in this direction, there
      // is some trickery involved in solving the system in this direction.
      // The first required thing is to solve two separate systems
      triDiagonalSolver( x, Bx, Ax, Cx, bx1, LENGTH-1 );
      triDiagonalSolver( y, Bx, Ax, Cx, bx2, LENGTH-1 );

      // Now calculate the last element:
      hn[LENGTH-1][j] = ( bx1[LENGTH-1] - Cx[LENGTH-1]*x[0] - Bx[LENGTH-1]*x[LENGTH-2] )/( Ax[LENGTH-1] + Cx[LENGTH-1]*y[0] + Bx[LENGTH-1]*y[LENGTH-2] );

      // Put the solution back in the array
      for ( i=0; i < LENGTH-1; i++ ) { hn[i][j] = x[i] + y[i]*hn[LENGTH-1][j]; }
    }
    
    // Insert the lower boundary condition
    for ( i = sed_row_lower; i < sed_row_upper; i ++ ) { hn[i][0] = WATER_LEVEL; }

    // Now collect the results, apply the boundary condition,
    // broadcast the new values of H, and repeat for the next direction
    MPI_Allreduce( hn, H, LENGTH*WIDTH, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    // Be considerate and clear the array
    memset( hn, 0, LENGTH*WIDTH*sizeof(double) );

    /***************************************************************************
     * END: Sediment Flow Process
     **************************************************************************/

    // Now that the system has been eroded slightly, have The Master
    // check to see if the target percentage has been reached.
    // Find the current volume of sediment
    Hy = 0;
    for ( i = rank; i < LENGTH; i += size )
    {
      for ( j=0; j < WIDTH; j ++ )
      {
        Hy += ( H[i][j] - WATER_LEVEL )*DX*DY;
      }
    }
    MPI_Allreduce( &Hy, &cur_percent, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );

    if ( master )
    {
      // This converts the volume into the percent eroded.
      cur_percent = 100.0*( (volume - cur_percent)/volume );
    
      // Check if this hits the target
      if ( cur_percent >= target )
      {
        // Create the file name to save to
        oss <<path_name << "/frame" << index << ".txt";
        // Save the data to this file
        write_data( oss.str(), index, target, volume, H, h );       
        // Update the index and target
        index ++;
        target += PERCENT_STEP;
        // Save this data to the original file it was read from.
        // This way the code can pick up from where it left off, just in case.
        write_data( file_name, index, target, volume, H, h );

        // Determine if we have sufficiently eroded the system.
        done = ( target > GOAL );
	
        // Clear the stream
        oss.str("");      
      } // End Target Percent If
    } //End The Master's work

    // Have The Master broadcast the value of done
    MPI_Bcast( &done, 1, MPI::BOOL, 0, MPI_COMM_WORLD );
 
  } // End Erosion Process While  

  // End the Parallelism
  MPI_Finalize();

  // We have finished the erosion of this system.
  return 0;
} // End Main

