/******************************************************************************
 *
 * The function triDiagonalSolver  below will solve a tridiagonal system of 
 * linear equations by using the Thomas reduction algorithm. 
 * The system to solve is
 *				Ax = b
 * where A is a tridiagonal matrix and B, A, C, are the non zero diagonals
 * (lower, middle, upper, respectively).
 * The array b is the solution, and length of the matrices.
 *
 *****************************************************************************/
#include "triDiagonalSolver.h"

int triDiagonalSolver(double* x, double* B, double* A, double* C, double*b, int length ) {

  // Typically, solve a tridiagonal system in this way involves solving
  // for two triangular matricies: one lower and one upper triangular
  // matrix. Like with the matrix A, we only need three non zero 
  // diagonals. Create them now.
  double L [length];
  double D [length];
  double U [length];
	
  // Calculate the initial values of D and U
  D[0] = A[0];
  U[0] = C[0];
  
  // Find the rest of the values
  for ( int i = 1; i < length-1; i ++ ) {
    L[i] = B[i]/D[i-1];
    D[i] = A[i] - L[i]*U[i-1];
    U[i] = C[i];
  }

  // Calculate the last values
  L[length-1] = B[length-1]/D[length-2];
  D[length-1] = A[length-1] - L[length-1]*U[length-2];

  // Now that the nonzero values of L, D, and U have been calculated, we 
  // solve the linear system Lz = b. First we have to create the 
  // variable for z.
  double z [length];
  // Solve for z
  z[0] = b[0];
  for ( int j = 1; j < length; j ++ ) {
    z[j] = b[j] - L[j]*z[j-1];
  }	

  // Now solve Ux = z. 
  x[length-1] = z[length-1]/D[length-1];
  for ( int k = length-2; k >=0; k -- ) {
    x[k] = (z[k] - U[k]*x[k+1])/D[k];
  }

  // Now x is the solution to the original problem Ax = b and thus we are
  // done.
  return 1;
}
