function convertTextToGraphs(dataDir, graphDir, num, grade)
% This function will take all *.txt files from the dataDir directory and
% graph the data in various ways.

% Add a color for the land surface
land = [205/255, 133/255, 63/255];

% To make the 'Landscape' graphs, we need a gradient from deep water to no
% water. We create that gradient now:
blue = [0, 0, 1];
cyan = [0, 1, 1];

N = 100; % The number of colors to interpolate

surface = zeros(2*N+1, 3);
surface(1, :) = blue;
surface(N+1, :) = cyan;
surface(2*N+1, :) = land;

% Perform the actual interpolation
for c = 2:1:N
    surface(c, :) = (c-1)*( (cyan - blue)/(N-1) ) + blue;
    surface(N+c, :) = (c-1)*( (land - cyan)/(N-1) ) + cyan;
end
num_colors = length(surface);
little_water = 0;
deep_water = 0.05;
C_Map = zeros(101, 101, 3);
interval_length = (deep_water - little_water)/(num_colors - 1);

% Now majority of the prep work is done. Now get the data and graph it.
dataFile = dir( fullfile( dataDir, '*.txt' ) );

% Graph each data file
files = length( dataFile );

for f = 1:files
    % Load the data
    data = strcat( dataDir, '/', dataFile(f).name );
    [index, ~, ~, H, h] = read_data( data, 101, 101 );
    
    % Create the color map now as this is used for two of the three graphs
    for j = 1:1:101
        for i = 1:1:101
            if ( h(i, j) <= little_water )
                C_Map(j, i, :) = land;
            elseif ( h(i, j) > deep_water )
                C_Map(j, i, :) = blue;
            else
                val = ceil( (h(i,j)-little_water)/interval_length );
                C_Map(j, i, :) = surface(num_colors-val, :);
            end
        end
    end
    
    % Now create and save the appropriate graphs:
    % if num is even, make the usual surface graph
    % if num is divisible by 3, make the landscape graph
    % if num is divisible by 5, make the water height graph
    if ( mod(num, 2) == 0 )
        % Print the surface
        % Create the figure and set its properties
        fig0 = figure('visible', 'off');
        axes0 = axes('Parent', fig0);
        zMax = 10 + 2*ceil((grade-4)/2);
        zlim([4, zMax]);
        set( axes0, 'ZTick', linspace(4, zMax, (zMax-4)/2 + 1) );
        set(axes0, 'YTick',[0 100]);
        set(axes0, 'XTick',[0 100]);
        zlabel('Height (meters)');
        view(axes0, [-31.5 36]);
        grid(axes0,'on');
        hold(axes0,'all');
        
        % Plot the surface
        surf(0:1:100, 0:1:100, H', 'FaceColor', land, 'EdgeColor', 'none');
        hold on;
        % Change the lighting and the surface material
        camlight left;
        material dull;
        lighting gouraud;
        % Plot the sides
        patch( [0, 0:1:100, 100], zeros(103, 1), [4, H(:, 1)', 4], land, 'EdgeColor', 'none');
        patch( zeros(103, 1), [0:1:100, 100, 0], [H(1, :), 4, 4], land, 'EdgeColor', 'none' );
        hold off;
        
        % Save and close the figure
        name = sprintf('/Surface/frame%d', index);
        print(fig0, '-djpeg', strcat(graphDir, name));
        close(fig0);
    end
    if ( mod(num, 3) == 0 )
        % Print the Landscape
        % Create the figure and set its properties 
        fig1 = figure('visible', 'off');
        axes1 = axes('Parent', fig1);
        zMax = 10 + 2*ceil((grade-4)/2);
        zlim([4, zMax]);
        set( axes0, 'ZTick', linspace(4, zMax, (zMax-4)/2 + 1) );
        set(axes1, 'YTick',[0 100]);
        set(axes1, 'XTick',[0 100]);
        zlabel('Height (meters)');
        view(axes1, [-31.5 36]);
        grid(axes1,'on');
        hold(axes1,'all');

        % Use the created colormap to plot the landscape
        surf(0:1:100, 0:1:100, H', C_Map, 'FaceColor', 'interp', 'EdgeColor', 'none');
        hold on;
        % Lighting and effects
        camlight left; 
        material dull;
        lighting gouraud;
        % Graph the sides
        patch( [0, 0:1:100, 100], zeros(103, 1), [4, H(:, 1)', 4], land, 'EdgeColor', 'none');
        patch( zeros(103, 1), [0:1:100, 100, 0], [H(1, :), 4, 4], land, 'EdgeColor', 'none' );
        hold off;
        % Save and close
        name = sprintf('/Landscape/frame%d', index);
        print(fig1, '-djpeg', strcat(graphDir, name));
        close(fig1);
    end
    if ( mod(num, 5) == 0 )
        % Print the water depth
        % Create the figure and set its properties
        fig2 = figure('visible', 'off');
        axes2 = axes('Parent', fig1);
        zlim([0, 0.1]);
        set(axes2, 'ZTick', [0 0.025 0.05 0.075 0.1]);
        set(axes2, 'YTick',[0 100]);
        set(axes2, 'XTick',[0 100]);
        view(axes2, [0.5 90]);
        grid(axes2,'on');
        hold(axes2,'all');
        
        % Graph the data
        surf(0:1:100, 0:1:100, h', C_Map, 'FaceColor', 'interp', 'EdgeColor', 'none');
        % Save and close
        name = sprintf('/Water/frame%d', index);
        print(fig2, '-djpeg', strcat(graphDir, name));
        close(fig2);
    end
    
    % All the data has been plotted. Clear the data
    clear H h index
    
    fprintf('Converted %d of %d\n', f, files);
end
    
    